"""
Time Axes Encoders
------------------

Encoders can generate past and/or future covariates series by encoding the index of a TimeSeries `series`.
Each encoder class has methods `encode_train()`, `encode_inference()`, and `encode_train_inference()` to generate the
encodings for training and inference.

The encoders extract the index either from the target series or optional additional past/future covariates.
If additional covariates are supplied to `encode_train()`, `encode_inference()`, or `encode_train_inference()`,
the time index of those covariates are used for the encodings. This means that the input covariates must meet the same
model-specific requirements as without encoders.

There are two main types of encoder classes: `SingleEncoder` and `SequentialEncoder`.

*   SingleEncoder
        The SingleEncoder classes carry the encoder logic for past and future covariates, and training and
        inference datasets. They can be used as stand-alone encoders.

        Each SingleEncoder has a dedicated subclass for generating past or future covariates. The naming convention
        is `{X}{SingleEncoder}` where {X} is one of (Past, Future) and {SingleEncoder} is one of the SingleEncoder
        classes described in the next section. An example:

        .. highlight:: python
        .. code-block:: python

            encoder = PastDatetimeAttributeEncoder(
                input_chunk_length=24,
                output_chunk_length=12,
                attribute='month'
                tz='CET'
            )

            past_covariates_train = encoder.encode_train(
                target=target,
                covariates=optional_past_covariates
            )
            past_covariates_inf = encoder.encode_inference(
                n=12,
                target=target,
                covariates=optional_past_covariates
            )
            # or generate encodings for train and inference together
            past_covariates_train_inf = encoder.encode_train_inference(
                n=12,
                target=target,
                covariates=optional_past_covariates
            )

*   SequentialEncoder
        Stores and controls multiple SingleEncoders for both past and/or future covariates all under one hood.
        It provides the same functionality as SingleEncoders (`encode_train()`, `encode_inference()`, and
        `encode_train_inference()`).
        It can be used both as stand-alone or as an all-in-one solution with Darts' forecasting models that support
        covariates through optional parameter `add_encoders`:

        .. highlight:: python
        .. code-block:: python

            model = SomeForecastingModel(..., add_encoders={...})
        ..

        If used at model creation, the SequentialEncoder will handle all past and future encoders autonomously.
        The requirements for model parameter `add_encoders` are described in the next section or in
        :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>`.

SingleEncoder
-------------

The SingleEncoders from {X}{SingleEncoder} are:

*   `DatetimeAttributeEncoder`
        Adds scalar pd.DatatimeIndex attribute information derived from `series.time_index`.
        Requires `series` to have a pd.DatetimeIndex.

        attribute
            An attribute of `pd.DatetimeIndex`: see all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
        tz
            Optionally, convert the time zone naive index to a time zone `tz` before applying the encoder.
*   `CyclicTemporalEncoder`
        Adds cyclic pd.DatetimeIndex attribute information deriveed from `series.time_index`.
        Adds 2 columns, corresponding to sin and cos encodings, to uniquely describe the underlying attribute.
        Requires `series` to have a pd.DatetimeIndex.

        attribute
            An attribute of `pd.DatetimeIndex` that follows a cyclic pattern. One of ('month', 'day', 'weekday',
            'dayofweek', 'day_of_week', 'hour', 'minute', 'second', 'microsecond', 'nanosecond', 'quarter',
            'dayofyear', 'day_of_year', 'week', 'weekofyear', 'week_of_year').
        tz
            Optionally, convert the time zone naive index to a time zone `tz` before applying the encoder.
*   `IntegerIndexEncoder`
        Adds the relative index positions as integer values (positions) derived from `series` time index.
        `series` can either have a pd.DatetimeIndex or an integer index.

        attribute
            Currently, only 'relative' is supported.
            'relative' will generate position values relative to the forecasting/prediction point. Values range
            from -inf to inf where 0 is set at the forecasting point.
*   `CallableIndexEncoder`
        Applies a user-defined callable to encode `series`' index.
        `series` can either have a pd.DatetimeIndex or an integer index.

        attribute
            a callable/function to encode the index.
            For `series` with a pd.DatetimeIndex: ``lambda index: (index.year - 1950) / 50``.
            For `series` with an integer index: ``lambda index: index / 50``

SequentialEncoder
-----------------

The SequentialEncoder combines the logic of all SingleEncoders from above and has additional benefits:

*   use multiple encoders at once
*   generate multiple attribute encodings at once
*   generate both past and future at once
*   supports transformers (Scaler)
*   easy to use with any forecasting model that supports covariates.

The model parameter `add_encoders` must be a Dict following of this convention:

*   outer keys: `SingleEncoder` and Transformer tags:

    *   'datetime_attribute' for `DatetimeAttributeEncoder`
    *   'cyclic' for `CyclicEncoder`
    *   'position' for `IntegerIndexEncoder`
    *   'custom' for `CallableIndexEncoder`
    *   'transformer' for a transformer
    *   'tz' for applying a time zone conversion
*   inner keys: covariates type

    *   'past' for past covariates
    *   'future' for future covariates
    *   (do not specify for 'transformer')
*   inner key values:

    *   list of attributes for `SingleEncoder`
    *   transformer object for 'transformer'

Below is an example that illustrates a valid `add_encoders` dict for hourly data and how it can be used with a
TorchForecastingModel (this is only meant to illustrate many features at once).

.. highlight:: python
.. code-block:: python

    add_encoders = {
        'cyclic': {'future': ['month']},
        'datetime_attribute': {'future': ['hour', 'dayofweek']},
        'position': {'past': ['relative'], 'future': ['relative']},
        'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
        'transformer': Scaler(),
        'tz': 'CET',
    }

    model = SomeTorchForecastingModel(..., add_encoders=add_encoders)
..
"""

import copy
from collections.abc import Sequence
from typing import Callable, Optional, Union

import numpy as np
import pandas as pd

from darts import TimeSeries, concatenate
from darts.dataprocessing.encoders.encoder_base import (
    CovariatesIndexGenerator,
    Encoder,
    FutureCovariatesIndexGenerator,
    PastCovariatesIndexGenerator,
    SequentialEncoderTransformer,
    SingleEncoder,
    SupportedIndex,
    _EncoderMethod,
)
from darts.dataprocessing.transformers import FittableDataTransformer
from darts.logging import get_logger, raise_if, raise_if_not
from darts.timeseries import DIMS
from darts.utils.timeseries_generation import datetime_attribute_timeseries
from darts.utils.ts_utils import seq2series, series2seq
from darts.utils.utils import generate_index

SupportedTimeSeries = Union[TimeSeries, Sequence[TimeSeries]]
logger = get_logger(__name__)

ENCODER_KEYS = ["cyclic", "datetime_attribute", "position", "custom"]
FUTURE = "future"
PAST = "past"
VALID_TIME_PARAMS = [FUTURE, PAST]
VALID_ENCODER_DTYPES = (str, Sequence)

TZ_KEYS = ["tz"]
TRANSFORMER_KEYS = ["transformer"]
VALID_TRANSFORMER_DTYPES = FittableDataTransformer
INTEGER_INDEX_ATTRIBUTES = ["relative"]


class CyclicTemporalEncoder(SingleEncoder):
    """`CyclicTemporalEncoder`: Cyclic encoding of time series datetime attributes."""

    def __init__(
        self,
        index_generator: CovariatesIndexGenerator,
        attribute: str,
        tz: Optional[str] = None,
    ):
        """
        Cyclic index encoding for `TimeSeries` that have a time index of type `pd.DatetimeIndex`.

        Parameters
        ----------
        index_generator
            An instance of `CovariatesIndexGenerator` with methods `generate_train_idx()` and
            `generate_inference_idx()`. Used to generate the index for encoders.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(index_generator)
        self.attribute = attribute
        self.tz = tz

    def _encode(
        self, index: SupportedIndex, target_end: pd.Timestamp, dtype: np.dtype
    ) -> TimeSeries:
        """applies cyclic encoding from `datetime_attribute_timeseries()` to `self.attribute` of `index`."""
        super()._encode(index, target_end, dtype)
        return datetime_attribute_timeseries(
            index,
            attribute=self.attribute,
            cyclic=True,
            dtype=dtype,
            with_columns=[
                self.base_component_name + self.attribute + "_sin",
                self.base_component_name + self.attribute + "_cos",
            ],
            tz=self.tz,
        )

    @property
    def accept_transformer(self) -> list[bool]:
        """`CyclicTemporalEncoder` should not be transformed. Returns two elements for sine and cosine waves."""
        return [False, False]

    @property
    def requires_fit(self) -> bool:
        return False

    @property
    def base_component_name(self) -> str:
        return super().base_component_name + "_cyc_"

    @property
    def encoding_n_components(self) -> int:
        return 2


class PastCyclicEncoder(CyclicTemporalEncoder):
    """`CyclicEncoder`: Cyclic encoding of past covariates datetime attributes."""

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        tz: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the past covariate lags. Accepts integer lag values <= -1.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_past_covariates` of :class:`RegressionModel`.
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(
            index_generator=PastCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
            tz=tz,
        )


class FutureCyclicEncoder(CyclicTemporalEncoder):
    """`CyclicEncoder`: Cyclic encoding of future covariates datetime attributes."""

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        tz: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the future covariate lags. Accepts all integer values.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_future_covariates` from :class:`RegressionModel`.
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(
            index_generator=FutureCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
            tz=tz,
        )


class DatetimeAttributeEncoder(SingleEncoder):
    """`DatetimeAttributeEncoder`: Adds pd.DatatimeIndex attribute information derived from the index as scalars.
    Requires the underlying TimeSeries to have a pd.DatetimeIndex
    """

    def __init__(
        self,
        index_generator: CovariatesIndexGenerator,
        attribute: str,
        tz: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariatesIndexGenerator` with methods `generate_train_idx()` and
            `generate_inference_idx()`. Used to generate the index for encoders.
        attribute
            The attribute of the underlying pd.DatetimeIndex for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(index_generator)
        self.attribute = attribute
        self.tz = tz

    def _encode(
        self, index: SupportedIndex, target_end: pd.Timestamp, dtype: np.dtype
    ) -> TimeSeries:
        """Encode `index` as a scalar."""
        super()._encode(index, target_end, dtype)
        return datetime_attribute_timeseries(
            index,
            attribute=self.attribute,
            dtype=dtype,
            with_columns=self.base_component_name + self.attribute,
            tz=self.tz,
        )

    @property
    def accept_transformer(self) -> list[bool]:
        """`DatetimeAttributeEncoder` accepts transformations"""
        return [True]

    @property
    def requires_fit(self) -> bool:
        return False

    @property
    def base_component_name(self) -> str:
        return super().base_component_name + "_dta_"

    @property
    def encoding_n_components(self) -> int:
        return 1


class PastDatetimeAttributeEncoder(DatetimeAttributeEncoder):
    """Datetime attribute encoder for past covariates."""

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        tz: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        attribute
            The attribute of the underlying pd.DatetimeIndex for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the past covariate lags. Accepts integer lag values <= -1.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_past_covariates` of :class:`RegressionModel`.
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(
            index_generator=PastCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
            tz=tz,
        )


class FutureDatetimeAttributeEncoder(DatetimeAttributeEncoder):
    """Datetime attribute encoder for future covariates."""

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        tz: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        attribute
            The attribute of the underlying pd.DatetimeIndex for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the future covariate lags. Accepts all integer values.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_future_covariates` from :class:`RegressionModel`.
        tz
            Optionally, a time zone to convert the time index to before computing the attributes.
        """
        super().__init__(
            index_generator=FutureCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
            tz=tz,
        )


class IntegerIndexEncoder(SingleEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) derived from the underlying TimeSeries' time index
    for past and future covariates.
    """

    def __init__(self, index_generator: CovariatesIndexGenerator, attribute: str):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariatesIndexGenerator` with methods `generate_train_idx()` and
            `generate_inference_idx()`. Used to generate the index for encoders.
        attribute
            Currently only 'relative' is supported. The generated encoded values will range from (-inf, inf) and the
            target series end time will be used as a reference to evaluate the relative index positions.
        """
        raise_if_not(
            isinstance(attribute, str) and attribute in INTEGER_INDEX_ATTRIBUTES,
            f"Encountered invalid encoder argument `{attribute}` for encoder `position`. "
            f'Attribute must be `"relative"`.',
            logger,
        )
        super().__init__(index_generator)

        self.attribute = attribute

    def _encode(
        self, index: SupportedIndex, target_end: pd.Timestamp, dtype: np.dtype
    ) -> TimeSeries:
        """Adds integer index value (position) to the provided `index`.
        For attribute=='relative', the reference point/index is the prediction/forecast index of the target series.
        """
        super()._encode(index, target_end, dtype)

        idx_larger_end = (index <= target_end).sum()
        freq = index.freq if isinstance(index, pd.DatetimeIndex) else index.step
        if idx_larger_end:
            idx_larger_end -= 1
        if index[0] > target_end:
            idx_diff = (
                len(generate_index(start=target_end, end=index[0], freq=freq)) - 1
            )
        elif index[-1] < target_end:
            idx_diff = (
                -len(generate_index(start=index[-1], end=target_end, freq=freq)) + 1
            )
        else:
            idx_diff = 0
        return TimeSeries.from_times_and_values(
            times=index,
            values=np.arange(
                start=idx_diff - idx_larger_end,
                stop=idx_diff - idx_larger_end + len(index),
            ),
            columns=[self.base_component_name + self.attribute],
        ).astype(np.dtype(dtype))

    @property
    def accept_transformer(self) -> list[bool]:
        """`IntegerIndexEncoder` accepts transformations. Note that transforming 'relative' `IntegerIndexEncoder`
        will return the absolute position (in the transformed space)."""
        return [True]

    @property
    def requires_fit(self) -> bool:
        # requires fitting to get the reference index from `IntegerIndexEncoder.index_generator` for inference
        return True

    @property
    def base_component_name(self) -> str:
        return super().base_component_name + "_pos_"

    @property
    def encoding_n_components(self) -> int:
        return 1


class PastIntegerIndexEncoder(IntegerIndexEncoder):
    """`IntegerIndexEncoder`: Adds integer index value (position) for past covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        attribute
            Currently only 'relative' is supported. The generated encoded values will range from (-inf, inf) and the
            target series end time will be used as a reference to evaluate the relative index positions.
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the past covariate lags. Accepts integer lag values <= -1.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_past_covariates` of :class:`RegressionModel`.
        """
        super().__init__(
            index_generator=PastCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
        )


class FutureIntegerIndexEncoder(IntegerIndexEncoder):
    """`IntegerIndexEncoder`: Adds integer index value (position) for future covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(
        self,
        attribute: str,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        attribute
            Currently only 'relative' is supported. The generated encoded values will range from (-inf, inf) and the
            target series end time will be used as a reference to evaluate the relative index positions.
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the future covariate lags. Accepts all integer values.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_future_covariates` from :class:`RegressionModel`.
        """
        super().__init__(
            index_generator=FutureCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
        )


class CallableIndexEncoder(SingleEncoder):
    """`CallableIndexEncoder`: Applies a user-defined callable to encode the underlying index for past and future
    covariates.
    """

    def __init__(self, index_generator: CovariatesIndexGenerator, attribute: Callable):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariatesIndexGenerator` with methods `generate_train_idx()` and
            `generate_inference_idx()`. Used to generate the index for encoders.
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `index` of type pd.DatetimeIndex:
            ``attribute = lambda index: (index.year - 1950) / 50``. And for pd.RangeIndex:
            ``attribute = lambda index: (index - 1950) / 50``
        """
        raise_if_not(
            callable(attribute),
            f"Encountered invalid encoder argument `{attribute}` for encoder `callable`. "
            f"Attribute must be a callable that returns a `np.ndarray`.",
            logger,
        )

        super().__init__(index_generator)

        self.attribute = attribute

    def _encode(
        self, index: SupportedIndex, target_end: pd.Timestamp, dtype: np.dtype
    ) -> TimeSeries:
        """Apply the user-defined callable to encode the index"""
        super()._encode(index, target_end, dtype)

        return TimeSeries.from_times_and_values(
            times=index,
            values=self.attribute(index),
            columns=[self.base_component_name + "custom"],
        ).astype(np.dtype(dtype))

    @property
    def accept_transformer(self) -> list[bool]:
        """`CallableIndexEncoder` accepts transformations."""
        return [True]

    @property
    def requires_fit(self) -> bool:
        return False

    @property
    def base_component_name(self) -> str:
        return super().base_component_name + "_cus_"

    @property
    def encoding_n_components(self) -> int:
        return 1


class PastCallableIndexEncoder(CallableIndexEncoder):
    """`IntegerIndexEncoder`: Adds integer index value (position) for past covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(
        self,
        attribute: Callable,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `index` of type pd.DatetimeIndex:
            ``attribute = lambda index: (index.year - 1950) / 50``. And for pd.RangeIndex:
            ``attribute = lambda index: (index - 1950) / 50``
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the past covariate lags. Accepts integer lag values <= -1.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_past_covariates` of :class:`RegressionModel`.
        """
        super().__init__(
            index_generator=PastCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
        )


class FutureCallableIndexEncoder(CallableIndexEncoder):
    """`IntegerIndexEncoder`: Adds integer index value (position) for future covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(
        self,
        attribute: Callable,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[list[int]] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `index` of type pd.DatetimeIndex:
            ``attribute = lambda index: (index.year - 1950) / 50``. And for pd.RangeIndex:
            ``attribute = lambda index: (index - 1950) / 50``
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers representing the future covariate lags. Accepts all integer values.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_future_covariates` from :class:`RegressionModel`.
        """
        super().__init__(
            index_generator=FutureCovariatesIndexGenerator(
                input_chunk_length,
                output_chunk_length,
                lags_covariates=lags_covariates,
            ),
            attribute=attribute,
        )


class SequentialEncoder(Encoder):
    """A `SequentialEncoder` object can store and control multiple past and future covariates encoders at once.
    It provides the same functionality as single encoders (`encode_train()`, `encode_inference()`,
    `encode_train_inference()`).
    """

    def __init__(
        self,
        add_encoders: dict,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_past_covariates: Optional[list[int]] = None,
        lags_future_covariates: Optional[list[int]] = None,
        takes_past_covariates: bool = False,
        takes_future_covariates: bool = False,
    ) -> None:
        """
        SequentialEncoder automatically creates encoder objects from parameter `add_encoders`. `add_encoders` can also
        be set directly in all of Darts' `ForecastingModels`. This will automatically set up a
        :class:`SequentialEncoder` tailored to the settings of the underlying forecasting model.

        The `add_encoders` dict must follow this convention:
            `{encoder keyword: {temporal keyword: List[attributes]}, ..., transformer keyword: transformer object}`
        Supported encoder keywords:
            `'cyclic'` for cyclic temporal encoder. See the docs
            :meth:`CyclicTemporalEncoder <darts.dataprocessing.encoders.CyclicTemporalEncoder>`;
            `'datetime_attribute'` for adding scalar information of pd.DatetimeIndex attribute. See the docs
            :meth:`DatetimeAttributeEncoder <darts.dataprocessing.encoders.DatetimeAttributeEncoder>`
            `'position'` for integer index position encoder. See the docs
            :meth:`IntegerIndexEncoder <darts.dataprocessing.encoders.IntegerIndexEncoder>`;
            `'custom'` for encoding index with custom callables (functions). See the docs
            :meth:`CallableIndexEncoder <darts.dataprocessing.encoders.CallableIndexEncoder>`;
        Supported temporal keywords:
            'past' for adding encoding as past covariates
            'future' for adding encoding as future covariates
        Supported attributes:
            for attributes read the referred docs for the corresponding encoder from above
        Supported transformers:
            a transformer can be added with transformer keyword 'transformer'. The transformer object must be an
            instance of Darts' :meth:`FittableDataTransformer
            <darts.dataprocessing.transformers.fittable_data_transformer.FittableDataTransformer>` such as Scaler() or
            BoxCox(). The transformers will be fitted on the training dataset when calling calling `model.fit()`.
            The training, validation and inference datasets are then transformed equally.
        Supported time zone:
            Optionally, apply a time zone conversion with keyword 'tz'. This converts the time zone-naive index to a
            timezone `'tz'` before applying the `'cyclic'` or `'datetime_attribute'` temporal encoders.

        An example of a valid `add_encoders` dict for hourly data:

            .. highlight:: python
            .. code-block:: python

                from darts.dataprocessing.transformers import Scaler
                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'past': ['hour'], 'future': ['year', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler(),
                    'tz': 'CET',
                }

        Tuples of `(encoder_id, attribute)` are extracted from `add_encoders` to instantiate the `SingleEncoder`
        objects:

        * The `encoder_id` is extracted as follows:
            str(encoder_kw) + str(temporal_kw) -> 'cyclic' + 'past' -> `encoder_id` = 'cyclic_past'
            The `encoder_id` is used to map the parameters with the corresponding `SingleEncoder` objects.
        * The `attribute` is extracted from the values given by values under `temporal_kw`
            `attribute` = 'month'
            ...
            The `attribute` tells the `SingleEncoder` which attribute of the index to encode

        New encoders can be added by appending them to the mapping property `SequentialEncoder.encoder_map()`

        Parameters
        ----------
        add_encoders
            A dictionary with the encoder settings.
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `input_chunk_length` from :class:`TorchForecastingModel`, or to the absolute
            minimum target lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required for
            :class:`TorchForecastingModel`, and :class:`RegressionModel`.
            Corresponds to parameter `output_chunk_length` from both :class:`TorchForecastingModel`, and
            :class:`RegressionModel`.
        lags_past_covariates
            Optionally, a list of integers representing the past covariate lags. Accepts integer lag values <= -1.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_past_covariates` of :class:`RegressionModel`.
        lags_future_covariates
            Optionally, a list of integers representing the future covariate lags. Accepts all integer values.
            Only required for :class:`RegressionModel`.
            Corresponds to the lag values from parameter `lags_future_covariates` from :class:`RegressionModel`.
        takes_past_covariates
            Whether to encode/generate past covariates.
        takes_future_covariates
            Whether to encode/generate future covariates.
        """
        super().__init__()
        self.params = add_encoders
        self.input_chunk_length = input_chunk_length
        self.output_chunk_length = output_chunk_length
        self.encoding_available = False
        self.takes_past_covariates = takes_past_covariates
        self.takes_future_covariates = takes_future_covariates
        self.lags_past_covariates = lags_past_covariates
        self.lags_future_covariates = lags_future_covariates

        # encoders
        self._past_encoders: list[SingleEncoder] = []
        self._past_components: pd.Index = pd.Index([])
        self._future_encoders: list[SingleEncoder] = []
        self._future_components: pd.Index = pd.Index([])

        # transformer
        self._past_transformer: Optional[SequentialEncoderTransformer] = None
        self._future_transformer: Optional[SequentialEncoderTransformer] = None

        # setup encoders and transformer
        self._setup_encoders(self.params)
        self._setup_transformer(self.params)

    def encode_train(
        self,
        target: SupportedTimeSeries,
        past_covariates: Optional[SupportedTimeSeries] = None,
        future_covariates: Optional[SupportedTimeSeries] = None,
        encode_past: bool = True,
        encode_future: bool = True,
    ) -> tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        """Returns encoded index for all past and/or future covariates for training.
        Which covariates are generated depends on the parameters used at model creation.

        Parameters
        ----------
        target
            The target TimeSeries used during training or passed to prediction as `series`.
        past_covariates
            Optionally, the past covariates used for training.
        future_covariates
            Optionally, the future covariates used for training.
        encode_past
            Whether to apply encoders for past covariates.
        encode_future
            Whether to apply encoders for future covariates.
        Returns
        -------
        Tuple[past_covariates, future_covariates]
            The past_covariates and/or future_covariates for training including the encodings.
            If input {x}_covariates is None and no {x}_encoders are given, will return `None`
            for the {x}_covariates.
        Raises
        ------
        Warning
            If model was created with `add_encoders` and there is suspicion of lazy loading.
            The encodings/covariates are generated eagerly before starting training for all individual targets and
            loaded into memory. Depending on the size of target data, this can create memory
            issues. In case this applies, consider setting `add_encoders=None` at model
            creation and build your encodings covariates manually for lazy loading.
        """
        if not self.fit_called:
            if not isinstance(target, (TimeSeries, list)):
                logger.warning(
                    "Fitting was called with `add_encoders` and suspicion of lazy loading. "
                    "The encodings/covariates are generated pre-train for all individual targets and "
                    "loaded into memory. Depending on the size of your data, this can create memory issues. "
                    "In case this applies, consider setting `add_encoders=None` at model creation."
                )

            self._fit_called = True
        past_covariates, future_covariates = self._launch_encoder(
            target=target,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            encoder_method=_EncoderMethod("train"),
            n=None,
            encode_past=encode_past,
            encode_future=encode_future,
        )
        self._fit_called = True
        return past_covariates, future_covariates

    def encode_inference(
        self,
        n: int,
        target: SupportedTimeSeries,
        past_covariates: Optional[SupportedTimeSeries] = None,
        future_covariates: Optional[SupportedTimeSeries] = None,
        encode_past: bool = True,
        encode_future: bool = True,
    ) -> tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        """Returns encoded index for all past and/or future covariates for inference/prediction.
        Which covariates are generated depends on the parameters used at model creation.

        Parameters
        ----------
        n
            The forecast horizon
        target
            The target TimeSeries used during training or passed to prediction as `series`.
        past_covariates
            Optionally, the past covariates used for training.
        future_covariates
            Optionally, the future covariates used for training.
        encode_past
            Whether to apply encoders for past covariates.
        encode_future
            Whether to apply encoders for future covariates.

        Returns
        -------
        Tuple[past_covariates, future_covariates]
            The past_covariates and/or future_covariates for prediction/inference including the encodings.
            If input {x}_covariates is None and no {x}_encoders are given, will return `None`
            for the {x}_covariates.
        """
        raise_if(
            not self.fit_called and self.requires_fit,
            f"`{self.__class__.__name__}` contains encoders or transformers which must be trained before inference. "
            "Call method `encode_train()` before `encode_inference()`.",
            logger=logger,
        )
        return self._launch_encoder(
            target=target,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            encoder_method=_EncoderMethod("inference"),
            n=n,
            encode_past=encode_past,
            encode_future=encode_future,
        )

    def encode_train_inference(
        self,
        n: int,
        target: SupportedTimeSeries,
        past_covariates: Optional[SupportedTimeSeries] = None,
        future_covariates: Optional[SupportedTimeSeries] = None,
        encode_past: bool = True,
        encode_future: bool = True,
    ) -> tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        """Returns encoded index for all past and/or future covariates for training and inference/prediction.
        Which covariates are generated depends on the parameters used at model creation.

        Parameters
        ----------
        n
            The forecast horizon
        target
            The target TimeSeries used for training and prediction.
        past_covariates
            Optionally, the past covariates used for training and prediction.
        future_covariates
            Optionally, the future covariates used for training and prediction.
        encode_past
            Whether to apply encoders for past covariates.
        encode_future
            Whether to apply encoders for future covariates.

        Returns
        -------
        Tuple[past_covariates, future_covariates]
            The past_covariates and/or future_covariates for prediction/inference including the encodings.
            If input {x}_covariates is None and no {x}_encoders are given, will return `None`
            for the {x}_covariates.
        """
        if not self.fit_called:
            if not isinstance(target, (TimeSeries, list)):
                logger.warning(
                    "Fitting was called with `add_encoders` and suspicion of lazy loading. "
                    "The encodings/covariates are generated pre-train for all individual targets and "
                    "loaded into memory. Depending on the size of your data, this can create memory issues. "
                    "In case this applies, consider setting `add_encoders=None` at model creation."
                )

            self._fit_called = True
        past_covariates, future_covariates = self._launch_encoder(
            target=target,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            encoder_method=_EncoderMethod("train_inference"),
            n=n,
            encode_past=encode_past,
            encode_future=encode_future,
        )
        self._fit_called = True
        return past_covariates, future_covariates

    def _launch_encoder(
        self,
        target: Sequence[TimeSeries],
        past_covariates: SupportedTimeSeries,
        future_covariates: SupportedTimeSeries,
        encoder_method: _EncoderMethod,
        n: Optional[int] = None,
        encode_past: bool = True,
        encode_future: bool = True,
    ) -> tuple[Sequence[TimeSeries], Sequence[TimeSeries]]:
        """Launches the encode sequence for past covariates and future covariates for either training,
        inference/prediction or training and inference/prediction depending on `encoder_method`.
        """
        if not self.encoding_available:
            return past_covariates, future_covariates

        # guarantee that all inputs are either a sequence of TimeSeries or None
        single_series = isinstance(target, TimeSeries)
        target = series2seq(target)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        # generate past covariates encodings
        if self.past_encoders and encode_past:
            past_covariates = self._encode_sequence(
                encoders=self.past_encoders,
                transformer=self.past_transformer,
                target=target,
                covariates=past_covariates,
                covariates_type=PAST,
                encoder_method=encoder_method,
                n=n,
            )

        # generate future covariates encodings
        if self.future_encoders and encode_future:
            future_covariates = self._encode_sequence(
                encoders=self.future_encoders,
                transformer=self.future_transformer,
                target=target,
                covariates=future_covariates,
                covariates_type=FUTURE,
                encoder_method=encoder_method,
                n=n,
            )

        # convert covariates back to single series if single target was used as input
        if single_series:
            past_covariates = seq2series(past_covariates)
            future_covariates = seq2series(future_covariates)
        return past_covariates, future_covariates

    def _encode_sequence(
        self,
        encoders: Sequence[SingleEncoder],
        transformer: Optional[SequentialEncoderTransformer],
        target: Sequence[TimeSeries],
        covariates: Optional[SupportedTimeSeries],
        covariates_type: str,
        encoder_method: _EncoderMethod,
        n: Optional[int] = None,
    ) -> list[TimeSeries]:
        """Sequentially encodes the index of all input target/covariates TimeSeries with the corresponding
        `encoder_method`.
        """
        encode_method = encoder_method.method

        encoded_sequence = []
        if covariates is None:
            covariates = [None] * len(target)
        else:
            covariates = (
                [covariates] if isinstance(covariates, TimeSeries) else covariates
            )

        for ts, covs in zip(target, covariates):
            # drop encoder components if they are in input covariates
            covs = self._drop_encoded_components(
                covariates=covs,
                components=getattr(self, f"{covariates_type}_components"),
            )
            encoded = concatenate(
                [
                    getattr(enc, encode_method)(
                        target=ts, covariates=covs, merge_covariates=False, n=n
                    )
                    for enc in encoders
                ],
                axis=DIMS[1],
            )
            encoded_sequence.append(
                self._merge_covariates(encoded=encoded, covariates=covs)
            )

        if transformer is not None:
            encoded_sequence = transformer.transform(encoded_sequence)

        # store encoded past/future component names if they were not saved before
        if getattr(self, f"{covariates_type}_components").empty:
            components = encoded_sequence[0].components
            if covariates is not None and covariates[0] is not None:
                components = components[~components.isin(covariates[0].components)]
            setattr(self, f"_{covariates_type}_components", components)

        return encoded_sequence

    @property
    def past_encoders(self) -> list[SingleEncoder]:
        """Returns the past covariates encoders"""
        return self._past_encoders

    @property
    def future_encoders(self) -> list[SingleEncoder]:
        """Returns the future covariates encoders"""
        return self._future_encoders

    @property
    def encoders(self) -> tuple[list[SingleEncoder], list[SingleEncoder]]:
        """Returns a tuple of (past covariates encoders, future covariates encoders)"""
        return self.past_encoders, self.future_encoders

    @property
    def past_components(self) -> pd.Index:
        """Returns the past covariates component names generated by `SequentialEncoder.past_encoders`.
        Only available after calling `SequentialEncoder.encode_train()`
        """
        return self._past_components

    @property
    def future_components(self) -> pd.Index:
        """Returns the future covariates component names generated by `SequentialEncoder.future_encoders`.
        Only available after calling `SequentialEncoder.encode_train()`
        """
        return self._future_components

    @property
    def components(self) -> tuple[pd.Index, pd.Index]:
        """Returns the covariates component names generated by `SequentialEncoder.past_encoders` and
        `SequentialEncoder.future_encoders`. A tuple of (past encoded components, future encoded components).
        Only available after calling `SequentialEncoder.encode_train()`
        """
        return self.past_components, self.future_components

    @property
    def encoding_n_components(self) -> tuple[int, int]:
        """Returns the number of components generated by `SequentialEncoder.past_encoders` and
        `SequentialEncoder.future_encoders`.
        """
        # by default, _[past/future]_encoders are empty lists
        past_enc_n_compoments = sum(
            past_enc.encoding_n_components for past_enc in self.past_encoders
        )
        future_enc_n_compoments = sum(
            future_enc.encoding_n_components for future_enc in self.future_encoders
        )
        return past_enc_n_compoments, future_enc_n_compoments

    @property
    def past_transformer(self) -> SequentialEncoderTransformer:
        """Returns the past transformer object"""
        return self._past_transformer

    @property
    def future_transformer(self) -> SequentialEncoderTransformer:
        """Returns the future transformer object"""
        return self._future_transformer

    def transformers(
        self,
    ) -> tuple[SequentialEncoderTransformer, SequentialEncoderTransformer]:
        """Returns a tuple of (past transformer, future transformer)."""
        return self.past_transformer, self.future_transformer

    @property
    def encoder_map(self) -> dict:
        """Mapping between encoder identifier string (from parameters at model creations) and the corresponding
        future or past covariates encoder"""
        mapper = {
            "cyclic_past": PastCyclicEncoder,
            "cyclic_future": FutureCyclicEncoder,
            "datetime_attribute_past": PastDatetimeAttributeEncoder,
            "datetime_attribute_future": FutureDatetimeAttributeEncoder,
            "position_past": PastIntegerIndexEncoder,
            "position_future": FutureIntegerIndexEncoder,
            "custom_past": PastCallableIndexEncoder,
            "custom_future": FutureCallableIndexEncoder,
        }
        return mapper

    def _setup_encoders(self, params: dict) -> None:
        """Sets up/Initializes all past and future encoders and an optional transformer from `add_encoder` parameter
        used at model creation.


        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={'cyclic': {'past': ['month', 'dayofweek', ...], 'future': [same as for 'past']}}
        """
        past_encoders, future_encoders = self._process_input_encoders(params)
        tz = self._process_timezone(params)

        if not past_encoders and not future_encoders:
            return

        self._past_encoders = [
            self.encoder_map[enc_id](
                attribute=attr,
                input_chunk_length=self.input_chunk_length,
                output_chunk_length=self.output_chunk_length,
                lags_covariates=self.lags_past_covariates,
                tz=tz,
            )
            for enc_id, attr in past_encoders
        ]
        self._future_encoders = [
            self.encoder_map[enc_id](
                attribute=attr,
                input_chunk_length=self.input_chunk_length,
                output_chunk_length=self.output_chunk_length,
                lags_covariates=self.lags_future_covariates,
                tz=tz,
            )
            for enc_id, attr in future_encoders
        ]
        self.encoding_available = True

    def _setup_transformer(self, params: dict) -> None:
        """Sets up/Initializes an optional transformer from `add_encoder` parameter used at model creation.

        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={..., 'transformer': Scaler()}
        """
        (
            transformer,
            transform_past_mask,
            transform_future_mask,
        ) = self._process_input_transformer(params)
        if transform_past_mask:
            self._past_transformer = SequentialEncoderTransformer(
                copy.deepcopy(transformer), transform_past_mask
            )
        if transform_future_mask:
            self._future_transformer = SequentialEncoderTransformer(
                copy.deepcopy(transformer), transform_future_mask
            )

    def _process_input_encoders(self, params: dict) -> tuple[list, list]:
        """Processes input and returns two lists of tuples `(encoder_id, attribute)` from relevant encoder
        parameters at model creation.

        Parameters
        ----------
        params
            The `add_encoders` dict used at model creation. Must follow this convention:
                `{encoder keyword: {temporal keyword: List[attributes]}}`

            Tuples of `(encoder_id, attribute)` are extracted from `add_encoders` to instantiate the `SingleEncoder`
            objects:

            * The `encoder_id` is extracted as follows:
                str(encoder_kw) + str(temporal_kw) -> 'cyclic' + 'past' -> `encoder_id` = 'cyclic_past'
                The `encoder_id` is used to map the parameters with the corresponding `SingleEncoder` objects.
            * The `attribute` is extracted from the values given by values under `temporal_kw`
                `attribute` = 'month'
                ...
                The `attribute` tells the `SingleEncoder` which attribute of the index to encode

        Raises
        ------
        ValueError
            1) if the outermost key is other than (`past`, `future`)
            2) if the innermost values are other than type `str` or `Sequence`
            3) if any of entry in the innermost values is a lambda function
        """
        if not params:
            return [], []

        # check input for invalid encoder types
        invalid_encoders = [
            enc
            for enc in params
            if enc not in ENCODER_KEYS + TZ_KEYS + TRANSFORMER_KEYS
        ]
        raise_if(
            len(invalid_encoders) > 0,
            f"Encountered invalid encoder types `{invalid_encoders}` in `add_encoders` parameter at model "
            f"creation. Supported encoder types are: `{ENCODER_KEYS + TRANSFORMER_KEYS}`.",
            logger,
        )

        encoders = {enc: params[enc] for enc in ENCODER_KEYS if params.get(enc, None)}

        # check input for invalid temporal types
        invalid_time_params = list()
        for encoder, t_types in encoders.items():
            invalid_time_params += [
                t_type for t_type in t_types.keys() if t_type not in VALID_TIME_PARAMS
            ]

        raise_if(
            len(invalid_time_params) > 0,
            f"Encountered invalid temporal types `{invalid_time_params}` in `add_encoders` parameter at model "
            f"creation. Supported temporal types are: `{VALID_TIME_PARAMS}`.",
            logger,
        )

        # check that encoders are not lambda functions (not pickable)
        lambda_func_encoders = set()
        # convert into tuples of (encoder string identifier, encoder attribute)
        past_encoders, future_encoders = list(), list()
        for enc, enc_params in encoders.items():
            for enc_time, enc_attr in enc_params.items():
                raise_if_not(
                    isinstance(enc_attr, VALID_ENCODER_DTYPES),
                    f"Encountered value `{enc_attr}` of invalid type `{type(enc_attr)}` for encoder "
                    f"`{enc}` in `add_encoders` at model creation. Supported data types are: "
                    f"`{VALID_ENCODER_DTYPES}`.",
                    logger,
                )
                attrs = [enc_attr] if isinstance(enc_attr, str) else enc_attr
                for attr in attrs:
                    encoder_id = "_".join([enc, enc_time])
                    if enc_time == PAST:
                        past_encoders.append((encoder_id, attr))
                    else:
                        future_encoders.append((encoder_id, attr))

                    if isinstance(attr, Callable) and attr.__name__ == "<lambda>":
                        lambda_func_encoders.add(enc)

        raise_if(
            len(lambda_func_encoders) > 0,
            f"Encountered lambda function in the following `add_encoders` entries : {lambda_func_encoders} "
            f"at model creation. "
            f"In order to prevent issues when saving the model, these encoders must be converted to "
            f"named functions.",
            logger,
        )

        for temp_enc, takes_temp, temp in [
            (past_encoders, self.takes_past_covariates, "past"),
            (future_encoders, self.takes_future_covariates, "future"),
        ]:
            if temp_enc and not takes_temp:
                logger.warning(
                    f"Specified {temp} encoders in `add_encoders` at model creation but model does not "
                    f"accept {temp} covariates. {temp} encoders will be ignored."
                )

        past_encoders = past_encoders if self.takes_past_covariates else []
        future_encoders = future_encoders if self.takes_future_covariates else []
        return past_encoders, future_encoders

    def _process_input_transformer(
        self, params: dict
    ) -> tuple[Optional[FittableDataTransformer], list, list]:
        """Processes input params used at model creation and returns tuple of one transformer object and two masks
        that specify which past / future encoders accept being transformed.

        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={'transformer': Scaler()}
        """
        if not params:
            return None, [], []

        transformer = params.get(TRANSFORMER_KEYS[0], None)
        if transformer is None:
            return None, [], []

        raise_if_not(
            isinstance(transformer, VALID_TRANSFORMER_DTYPES),
            f"Encountered `{TRANSFORMER_KEYS[0]}` of invalid type `{type(transformer)}` "
            f"in `add_encoders` at model creation. Transformer must be an instance of "
            f"`{VALID_TRANSFORMER_DTYPES}`.",
            logger,
        )

        transform_past_mask = [
            transform
            for enc in self.past_encoders
            for transform in enc.accept_transformer
        ]
        transform_future_mask = [
            transform
            for enc in self.future_encoders
            for transform in enc.accept_transformer
        ]
        return transformer, transform_past_mask, transform_future_mask

    @staticmethod
    def _process_timezone(params: dict) -> Optional[str]:
        """Processes input params used at model creation for time zone specification, and returns the time zone.

        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={'tz': 'CET'}
        """
        if not params:
            return None

        return params.get(TZ_KEYS[0], None)

    @property
    def requires_fit(self) -> bool:
        return any([
            enc.requires_fit for cov_enc in self.encoders for enc in cov_enc
        ]) or any([tf is not None for tf in self.transformers()])



"""
Croston method
--------------
"""

from statsforecast.models import TSB as CrostonTSB
from statsforecast.models import CrostonClassic, CrostonOptimized, CrostonSBA

from darts.logging import raise_if, raise_if_not
from darts.models.forecasting.forecasting_model import (
    LocalForecastingModel,
)
from darts.timeseries import TimeSeries


class Croston(LocalForecastingModel):
    def __init__(
        self,
        version: str = "classic",
        alpha_d: float = None,
        alpha_p: float = None,
    ):
        """An implementation of the `Croston method
        <https://otexts.com/fpp3/counts.html>`_ for intermittent
        count series.

        Relying on the implementation of `Statsforecasts package
        <https://github.com/Nixtla/statsforecast>`_.

        Parameters
        ----------
        version
            - "classic" corresponds to classic Croston.
            - "optimized" corresponds to optimized classic Croston, which searches
              for the optimal ``alpha`` smoothing parameter and can take longer
              to run. Otherwise, a fixed value of ``alpha=0.1`` is used.
            - "sba" corresponds to the adjustment of the Croston method known as
              the Syntetos-Boylan Approximation [1]_.
            - "tsb" corresponds to the adjustment of the Croston method proposed by
              Teunter, Syntetos and Babai [2]_. In this case, `alpha_d` and `alpha_p` must
              be set.
        alpha_d
            For the "tsb" version, the alpha smoothing parameter to apply on demand.
        alpha_p
            For the "tsb" version, the alpha smoothing parameter to apply on probability.

        References
        ----------
        .. [1] Aris A. Syntetos and John E. Boylan. The accuracy of intermittent demand estimates.
               International Journal of Forecasting, 21(2):303 – 314, 2005.
        .. [2] Ruud H. Teunter, Aris A. Syntetos, and M. Zied Babai.
               Intermittent demand: Linking forecasting to inventory obsolescence.
               European Journal of Operational Research, 214(3):606 – 615, 2011.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.models import Croston
        >>> series = AirPassengersDataset().load()
        >>> # use the optimized version to automatically select best alpha parameter
        >>> model = Croston(version="optimized")
        >>> model.fit(series)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[461.7666],
               [461.7666],
               [461.7666],
               [461.7666],
               [461.7666],
               [461.7666]])
        """
        super().__init__(add_encoders=None)
        raise_if_not(
            version.lower() in ["classic", "optimized", "sba", "tsb"],
            'The provided "version" parameter must be set to "classic", "optimized", "sba" or "tsb".',
        )

        if version == "classic":
            self.model = CrostonClassic()
        elif version == "optimized":
            self.model = CrostonOptimized()
        elif version == "sba":
            self.model = CrostonSBA()
        else:
            raise_if(
                alpha_d is None or alpha_p is None,
                'alpha_d and alpha_p must be specified when using "tsb".',
            )
            self.alpha_d = alpha_d
            self.alpha_p = alpha_p
            self.model = CrostonTSB(alpha_d=self.alpha_d, alpha_p=self.alpha_p)

        self.version = version

    @property
    def supports_multivariate(self) -> bool:
        return False

    def fit(self, series: TimeSeries):
        super().fit(series)
        self._assert_univariate(series)
        series = self.training_series

        self.model.fit(
            y=series.values(copy=False).flatten(),
            # X can be used to passe future covariates only when conformal prediction is used
            X=None,
        )

        return self

    def predict(
        self,
        n: int,
        num_samples: int = 1,
        verbose: bool = False,
    ):
        super().predict(n, num_samples)
        values = self.model.predict(
            h=n,
            # X can be used to passe future covariates only when conformal prediction is used
            X=None,
        )["mean"]
        return self._build_forecast_series(values)

    @property
    def min_train_series_length(self) -> int:
        return 10

    @property
    def _supports_range_index(self) -> bool:
        return True



import itertools

import numpy as np
import pytest

from darts import TimeSeries
from darts.logging import get_logger
from darts.models import NaiveDrift, NaiveMean, NaiveMovingAverage, NaiveSeasonal
from darts.models.forecasting.forecasting_model import (
    GlobalForecastingModel,
    LocalForecastingModel,
)
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)


icl = 5
local_models = [
    (NaiveDrift, {}),
    (NaiveMean, {}),
    (NaiveMovingAverage, {}),
    (NaiveSeasonal, {}),
]
global_models = []


if TORCH_AVAILABLE:
    import torch

    from darts.models import GlobalNaiveAggregate, GlobalNaiveDrift, GlobalNaiveSeasonal

    global_models += [
        (
            GlobalNaiveAggregate,
            {"input_chunk_length": icl, "output_chunk_length": 3, **tfm_kwargs},
        ),
        (
            GlobalNaiveAggregate,
            {"input_chunk_length": icl, "output_chunk_length": 1, **tfm_kwargs},
        ),
        (
            GlobalNaiveDrift,
            {"input_chunk_length": icl, "output_chunk_length": 3, **tfm_kwargs},
        ),
        (
            GlobalNaiveDrift,
            {"input_chunk_length": icl, "output_chunk_length": 1, **tfm_kwargs},
        ),
        (
            GlobalNaiveSeasonal,
            {"input_chunk_length": icl, "output_chunk_length": 3, **tfm_kwargs},
        ),
        (
            GlobalNaiveSeasonal,
            {"input_chunk_length": icl, "output_chunk_length": 1, **tfm_kwargs},
        ),
    ]

    def custom_mean_valid(x, dim):
        return torch.mean(x, dim)

    def custom_mean_invalid_out_shape(x, dim):
        return x[:1]

    def custom_mean_invalid_signature(x):
        return torch.mean(x, dim=1)

    def custom_mean_invalid_output_type(x, dim):
        return torch.mean(x, dim=1).detach().numpy()

else:
    custom_mean_valid = None
    custom_mean_invalid_out_shape = None
    custom_mean_invalid_signature = None
    custom_mean_invalid_output_type = None


class TestBaselineModels:
    np.random.seed(42)
    if TORCH_AVAILABLE:
        torch.manual_seed(42)

    @pytest.mark.parametrize(
        "config", itertools.product(local_models + global_models, [False, True])
    )
    def test_fit_predict(self, config):
        """Tests fit and predict for univariate and multivariate time series."""
        (model_cls, model_kwargs), is_multivariate = config

        # min train series length for global naive models
        series = tg.linear_timeseries(length=icl)
        if is_multivariate:
            series.stack(series + 100)

        model = model_cls(**model_kwargs)

        # calling predict before fit
        with pytest.raises(ValueError):
            model.predict(n=10)

        # calling fit with covariates
        if isinstance(model, GlobalForecastingModel):
            err_type = ValueError
            err_msg_content = "The model does not support"
        else:  # for local models, covariates are not part of signature
            err_type = TypeError
            err_msg_content = "got an unexpected keyword argument"
        with pytest.raises(err_type) as err:
            model.fit(series=series, past_covariates=series)
        assert err_msg_content in str(err.value)
        with pytest.raises(err_type) as err:
            model.fit(series=series, future_covariates=series)
        assert err_msg_content in str(err.value)

        model.fit(series=series)
        # calling predict with covariates
        with pytest.raises(err_type) as err:
            model.predict(n=10, past_covariates=series)
        assert err_msg_content in str(err.value)
        with pytest.raises(err_type) as err:
            model.predict(n=10, future_covariates=series)
        assert err_msg_content in str(err.value)

        # single series predict works with all models
        preds = model.predict(n=10)
        preds_start = series.end_time() + series.freq
        assert isinstance(preds, TimeSeries)
        assert len(preds) == 10
        assert preds.start_time() == preds_start
        assert preds.components.equals(series.components)

        if isinstance(model, LocalForecastingModel):
            # no series at prediction time
            with pytest.raises(err_type) as err:
                _ = model.predict(n=10, series=series)
            assert err_msg_content in str(err.value)
            # no multiple series prediction
            with pytest.raises(err_type) as err:
                _ = model.predict(n=10, series=[series, series])
            assert err_msg_content in str(err.value)
        else:
            preds = model.predict(n=10, series=series)
            assert isinstance(preds, TimeSeries)
            assert len(preds) == 10
            assert preds.start_time() == preds_start
            assert preds.components.equals(series.components)
            preds = model.predict(n=10, series=[series, series])
            assert isinstance(preds, list)
            assert len(preds) == 2
            assert all([isinstance(p, TimeSeries) for p in preds])
            assert all([len(p) == 10 for p in preds])
            assert all([p.start_time() == preds_start for p in preds])
            assert all([p.components.equals(series.components) for p in preds])

        # multiple series training only with global baselines
        if isinstance(model, LocalForecastingModel):
            with pytest.raises(ValueError) as err:
                model.fit(series=[series, series])
            assert "Train `series` must be a single `TimeSeries`." == str(err.value)
        else:
            model.fit(series=[series, series])

    def test_naive_seasonal(self):
        # min train series length for global naive models
        series = tg.linear_timeseries(length=icl)
        series = series.stack(series + 25.0)

        vals_exp = series.values(copy=False)

        # local naive seasonal
        local_model = NaiveSeasonal(K=icl)
        preds = local_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        if not TORCH_AVAILABLE:
            return

        # equivalent global naive seasonal
        global_model = GlobalNaiveSeasonal(
            input_chunk_length=icl, output_chunk_length=1, **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), vals_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), vals_exp + 100.0
        )

        # global naive seasonal that repeats values `output_chunk_length` times
        global_model = GlobalNaiveSeasonal(
            input_chunk_length=icl, output_chunk_length=icl, **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(
            preds.values(copy=False), np.repeat(vals_exp[0:1, :], icl, axis=0)
        )

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), np.repeat(vals_exp[0:1, :], icl, axis=0)
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False),
            np.repeat(vals_exp[0:1, :] + 100.0, icl, axis=0),
        )

    def test_naive_drift(self):
        # min train series length for global naive models
        series_total = tg.linear_timeseries(length=2 * icl)
        series_total = series_total.stack(series_total + 25.0)
        series = series_total[:icl]
        series_drift = series_total[icl:]

        vals_exp = series_drift.values(copy=False)

        # local naive drift
        local_model = NaiveDrift()
        preds = local_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        if not TORCH_AVAILABLE:
            return

        # identical global naive drift
        global_model = GlobalNaiveDrift(
            input_chunk_length=icl, output_chunk_length=icl, **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), vals_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), vals_exp + 100.0
        )

        # global naive moving drift
        global_model = GlobalNaiveDrift(
            input_chunk_length=icl, output_chunk_length=1, **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)

        # manually compute the moving/autoregressive drift
        series_vals = series.values(copy=False)
        preds_vals = preds.values(copy=False)
        preds_exp = []
        x, y = 1, None
        for i in range(0, icl):
            y_0 = y if y is not None else series_vals[-1]
            m = (y_0 - series_vals[i]) / (icl - 1)
            y = m * x + y_0
            preds_exp.append(np.expand_dims(y, 0))
        preds_exp = np.concatenate(preds_exp)
        np.testing.assert_array_almost_equal(preds_vals, preds_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), preds_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), preds_exp + 100.0
        )

    def test_naive_mean(self):
        # min train series length for global naive models
        series = tg.linear_timeseries(length=icl)
        series = series.stack(series + 25.0)

        # mean repeated n times
        vals_exp = np.repeat(
            np.expand_dims(series.values(copy=False).mean(axis=0), 0), icl, axis=0
        )

        # local naive mean
        local_model = NaiveMean()
        preds = local_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        if not TORCH_AVAILABLE:
            return

        # identical global naive mean
        global_model = GlobalNaiveAggregate(
            input_chunk_length=icl, output_chunk_length=icl, agg_fn="mean", **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), vals_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), vals_exp + 100.0
        )

    def test_naive_moving_average(self):
        # min train series length for global naive models
        series = tg.linear_timeseries(length=icl)
        series = series.stack(series + 25.0)

        # manually compute the moving/autoregressive average/mean
        series_vals = series.values(copy=False)
        vals_exp = []
        y = None
        for i in range(0, icl):
            if y is None:
                y_moving = series_vals
            else:
                y_moving = np.concatenate(
                    [series_vals[i:], np.concatenate(vals_exp)], axis=0
                )
            y = np.expand_dims(y_moving.mean(axis=0), 0)
            vals_exp.append(y)
        vals_exp = np.concatenate(vals_exp)

        # local naive mean
        local_model = NaiveMovingAverage(input_chunk_length=icl)
        preds = local_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        if not TORCH_AVAILABLE:
            return

        # identical global naive moving average
        global_model = GlobalNaiveAggregate(
            input_chunk_length=icl, output_chunk_length=1, agg_fn="mean", **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), vals_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), vals_exp + 100.0
        )

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    @pytest.mark.parametrize(
        "agg_fn_config",
        [
            ("nanmean", "nanmean"),
            ("mean", "mean"),
            (custom_mean_valid, "mean"),
        ],
    )
    def test_global_naive_aggregate(self, agg_fn_config):
        agg_fn, agg_name = agg_fn_config

        # min train series length for global naive models
        series = tg.linear_timeseries(length=icl)
        series = series.stack(series + 25.0)

        # manually compute the moving/autoregressive average/mean
        series_vals = series.values(copy=False)
        vals_exp = []

        agg_fn_np = getattr(np, agg_name)
        y = None
        for i in range(0, icl):
            if y is None:
                y_moving = series_vals
            else:
                y_moving = np.concatenate(
                    [series_vals[i:], np.concatenate(vals_exp)], axis=0
                )

            y = np.expand_dims(agg_fn_np(y_moving, axis=0), 0)
            vals_exp.append(y)
        vals_exp = np.concatenate(vals_exp)

        # identical global naive moving average
        global_model = GlobalNaiveAggregate(
            input_chunk_length=icl, output_chunk_length=1, agg_fn=agg_fn, **tfm_kwargs
        )
        preds = global_model.fit(series).predict(n=icl)
        np.testing.assert_array_almost_equal(preds.values(copy=False), vals_exp)

        preds_multi = global_model.predict(n=icl, series=[series, series + 100.0])
        np.testing.assert_array_almost_equal(
            preds_multi[0].values(copy=False), vals_exp
        )
        np.testing.assert_array_almost_equal(
            preds_multi[1].values(copy=False), vals_exp + 100.0
        )

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    @pytest.mark.parametrize(
        "agg_fn_config",
        [
            ("mmean", "When `agg_fn` is a string"),
            (1, "`agg_fn` must be a string or callable"),
            (
                custom_mean_invalid_output_type,
                "`agg_fn` output must be a torch Tensor.",
            ),
            (custom_mean_invalid_signature, "got an unexpected keyword argument 'dim'"),
            (custom_mean_invalid_out_shape, "Unexpected `agg_fn` output shape."),
        ],
    )
    def test_global_naive_aggregate_invalid_agg_fn(self, agg_fn_config):
        agg_fn, err_msg_content = agg_fn_config
        # identical global naive moving average
        with pytest.raises(ValueError) as err:
            _ = GlobalNaiveAggregate(
                input_chunk_length=icl,
                output_chunk_length=1,
                agg_fn=agg_fn,
                **tfm_kwargs,
            )
        assert err_msg_content in str(err.value)



import numpy as np
import pandas as pd
import pytest
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder

from darts import TimeSeries
from darts.dataprocessing.transformers import StaticCovariatesTransformer
from darts.utils import timeseries_generation as tg


class TestStaticCovariatesTransformer:
    series = tg.linear_timeseries(length=10)
    static_covs1 = pd.DataFrame(
        data={
            "cont1": [0, 1, 2],
            "cat1": [1, 2, 3],
            "cont2": [0.1, 0.2, 0.3],
            "cat2": ["a", "b", "c"],
        }
    ).astype(dtype={"cat1": "O", "cat2": "O"})
    series1 = TimeSeries.from_times_and_values(
        times=series.time_index,
        values=np.concatenate([series.values()] * 3, axis=1),
        columns=["comp1", "comp2", "comp3"],
        static_covariates=static_covs1,
    )

    static_covs2 = pd.DataFrame(
        data={
            "cont1": [2, 3, 4],
            "cat1": [3, 4, 5],
            "cont2": [0.3, 0.4, 0.5],
            "cat2": ["c", "d", "e"],
        }
    )
    static_covs2["cat1"] = static_covs2["cat1"].astype("O")
    series2 = TimeSeries.from_times_and_values(
        times=series.time_index,
        values=np.concatenate([series.values()] * 3, axis=1),
        columns=["comp1", "comp2", "comp3"],
        static_covariates=static_covs2,
    )

    def test_scaling_single_series(self):
        # 3 categories for each categorical static covariate column (column idx 1 and 3)
        test_values = np.array([
            [0.0, 0.0, 0.0, 0.0],
            [0.5, 1.0, 0.5, 1.0],
            [1.0, 2.0, 1.0, 2.0],
        ])
        for series in [self.series1, self.series2]:
            scaler = StaticCovariatesTransformer()
            self.helper_test_scaling(series, scaler, test_values)

        test_values = np.array([
            [-1.0, 0.0, -1.0, 0.0],
            [0.0, 1.0, 0.0, 1.0],
            [1.0, 2.0, 1.0, 2.0],
        ])
        for series in [self.series1, self.series2]:
            scaler = StaticCovariatesTransformer(
                transformer_num=MinMaxScaler(feature_range=(-1, 1))
            )
            self.helper_test_scaling(series, scaler, test_values)

        test_values = np.array([
            [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.5, 0.0, 1.0, 0.0, 0.5, 0.0, 1.0, 0.0],
            [1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0],
        ])
        for series in [self.series1, self.series2]:
            scaler = StaticCovariatesTransformer(transformer_cat=OneHotEncoder())
            self.helper_test_scaling(series, scaler, test_values)

    def test_single_type_scaler(self):
        transformer_cont = StaticCovariatesTransformer()
        series_cont = self.series1.with_static_covariates(
            self.series1.static_covariates[["cont1", "cont2"]]
        )
        test_cont = np.array([[0.0, 0.0], [0.5, 0.5], [1.0, 1.0]])
        self.helper_test_scaling(series_cont, transformer_cont, test_cont)

        transformer_cat = StaticCovariatesTransformer()
        series_cat = self.series1.with_static_covariates(
            self.series1.static_covariates[["cat1", "cat2"]]
        )
        test_cat = np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]])
        self.helper_test_scaling(series_cat, transformer_cat, test_cat)

    def test_selected_columns(self):
        test_cont = (
            pd.DataFrame(
                [[0.0, 1, 0.0, "a"], [0.5, 2, 0.5, "b"], [1.0, 3, 1.0, "c"]],
            )
            .astype(dtype={1: "O", 3: "O"})
            .values
        )
        transformer_cont2 = StaticCovariatesTransformer(
            cols_num=["cont1", "cont2"], cols_cat=[]
        )
        self.helper_test_scaling(self.series1, transformer_cont2, test_cont)

        test_contcat = (
            pd.DataFrame(
                [[0.0, 1, 0.0, 0.0], [1.0, 2, 0.5, 1.0], [2.0, 3, 1.0, 2.0]],
            )
            .astype(dtype={1: "O"})
            .values
        )
        transformer_contcat = StaticCovariatesTransformer(
            cols_num=["cont2"], cols_cat=["cat2"]
        )
        self.helper_test_scaling(self.series1, transformer_contcat, test_contcat)

        test_cat = pd.DataFrame(
            [[0.0, 0.0, 0.1, 0.0], [1.0, 1.0, 0.2, 1], [2.0, 2.0, 0.3, 2.0]],
        ).values
        transformer_cat = StaticCovariatesTransformer(
            cols_num=[], cols_cat=["cat1", "cat2"]
        )
        self.helper_test_scaling(self.series1, transformer_cat, test_cat)

    def test_custom_scaler(self):
        # invalid scaler with missing inverse_transform
        class InvalidScaler:
            def fit(self):
                pass

            def transform(self):
                pass

        with pytest.raises(ValueError):
            _ = StaticCovariatesTransformer(transformer_num=InvalidScaler())

        with pytest.raises(ValueError):
            _ = StaticCovariatesTransformer(transformer_cat=InvalidScaler())

        class ValidScaler(InvalidScaler):
            def inverse_transform(self):
                pass

        _ = StaticCovariatesTransformer(transformer_num=ValidScaler())
        _ = StaticCovariatesTransformer(transformer_cat=ValidScaler())
        _ = StaticCovariatesTransformer(
            transformer_num=ValidScaler(), transformer_cat=ValidScaler()
        )

    def test_scaling_multi_series(self):
        # 5 categories in total for each categorical static covariate from multiple time series
        scaler = StaticCovariatesTransformer()
        series_tr2 = scaler.fit_transform([self.series1, self.series2])

        np.testing.assert_almost_equal(
            series_tr2[0].static_covariates_values(),
            np.array([
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 1.0, 0.25, 1.0],
                [0.5, 2.0, 0.5, 2.0],
            ]),
        )
        series_recovered2 = scaler.inverse_transform(series_tr2[0])
        assert self.series1.static_covariates.equals(
            series_recovered2.static_covariates
        )

        np.testing.assert_almost_equal(
            series_tr2[1].static_covariates_values(),
            np.array([
                [0.5, 2.0, 0.5, 2.0],
                [0.75, 3.0, 0.75, 3.0],
                [1.0, 4.0, 1.0, 4.0],
            ]),
        )
        series_recovered3 = scaler.inverse_transform(series_tr2[1])
        assert self.series2.static_covariates.equals(
            series_recovered3.static_covariates
        )

        series_recovered_multi = scaler.inverse_transform(series_tr2)
        assert self.series1.static_covariates.equals(
            series_recovered_multi[0].static_covariates
        )
        assert self.series2.static_covariates.equals(
            series_recovered_multi[1].static_covariates
        )

    def helper_test_scaling(self, series, scaler, test_values):
        series_tr = scaler.fit_transform(series)
        assert all([
            a == b
            for a, b in zip(
                series_tr.static_covariates_values().flatten(),
                test_values.flatten(),
            )
        ])

        series_recovered = scaler.inverse_transform(series_tr)
        assert series.static_covariates.equals(series_recovered.static_covariates)



"""
PyOD Scorer
-----

This scorer can wrap around detection algorithms of PyOD.
`PyOD https://pyod.readthedocs.io/en/latest/#`_.
"""

import numpy as np
from pyod.models.base import BaseDetector

from darts import metrics
from darts.ad.scorers.scorers import WindowedAnomalyScorer
from darts.logging import get_logger, raise_if_not
from darts.metrics.metrics import METRIC_TYPE

logger = get_logger(__name__)


class PyODScorer(WindowedAnomalyScorer):
    def __init__(
        self,
        model: BaseDetector,
        window: int = 1,
        component_wise: bool = False,
        window_agg: bool = True,
        diff_fn: METRIC_TYPE = metrics.ae,
    ) -> None:
        """PyOD Scorer

        When calling ``fit(series)``, a moving window is applied, which results in a set of vectors of size `W`,
        where `W` is the window size. The PyODScorer model is trained on these vectors. The ``score(series)``
        function will apply the same moving window and return the predicted raw anomaly score of each vector.

        Alternatively, the scorer has the functions ``fit_from_prediction()`` and ``score_from_prediction()``.
        Both require two series (actual and prediction), and compute a "difference" series by applying the
        function ``diff_fn`` (default: absolute difference). The resulting series is then passed to the
        functions ``fit()`` and ``score()``, respectively.

        `component_wise` is a boolean parameter indicating how the model should behave with multivariate inputs
        series. If set to `True`, the model will treat each series dimension independently by fitting a different
        PyODScorer model for each dimension. If set to `False`, the model concatenates the dimensions in
        each windows of length `W` and compute the score using only one underlying PyODScorer model.

        **Training with** ``fit()``:

        The input can be a series (univariate or multivariate) or multiple series. The series will be partitioned
        into equal size subsequences. The subsequence will be of size `W` * `D`, with:

        - `W` being the size of the window given as a parameter `window`
        - `D` being the dimension of the series (`D` = 1 if univariate or if `component_wise` is set to `True`)

        For a series of length `N`, (`N` - `W` + 1)/W subsequences will be generated. If a list of series is given
        of length L, each series will be partitioned into subsequences, and the results will be concatenated into
        an array of length L * number of subsequences of each series.

        The PyOD model will be fitted on the generated subsequences.

        If `component_wise` is set to `True`, the algorithm will be applied to each dimension independently. For each
        dimension, a PyOD model will be trained.

        **Computing score with** ``score()``:

        The input can be a series (univariate or multivariate) or a sequence of series. The given series must have the
        same dimension `D` as the data used to train the PyOD model.

        For each series, if the series is multivariate of dimension `D`:

        - if `component_wise` is set to `False`: it returns a univariate series (dimension=1). It represents
          the anomaly score of the entire series in the considered window at each timestamp.
        - if `component_wise` is set to `True`: it returns a multivariate series of dimension `D`. Each dimension
          represents the anomaly score of the corresponding component of the input.

        If the series is univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        A window of size `W` is rolled on the series with a stride equal to 1. It is the same size window `W` used
        during the training phase.
        Each value in the score series thus represents how anomalous the sample of the `W` previous values is.

        Parameters
        ----------
        model
            The (fitted) PyOD BaseDetector model.
        window
            Size of the window used to create the subsequences of the series.
        component_wise
            Boolean value indicating if the score needs to be computed for each component independently (`True`)
            or by concatenating the component in the considered window to compute one score (`False`).
            Default: `False`.
        window_agg
            Boolean indicating whether the anomaly score for each time step is computed by
            averaging the anomaly scores for all windows this point is included in.
            If `False`, the anomaly score for each point is the anomaly score of its trailing window.
            Default: `True`.
        diff_fn
            The differencing function to use to transform the predicted and actual series into one series.
            The scorer is then applied to this series. Must be one of Darts per-time-step metrics (e.g.,
            :func:`~darts.metrics.metrics.ae` for the absolute difference, :func:`~darts.metrics.metrics.err` for the
            difference, :func:`~darts.metrics.metrics.se` for the squared difference, ...).
            By default, uses the absolute difference (:func:`~darts.metrics.metrics.ae`).
        """

        raise_if_not(
            isinstance(model, BaseDetector),
            f"model must be a PyOD BaseDetector, found type: {type(model)}",
            logger,
        )
        self.model = model
        super().__init__(
            is_univariate=(not component_wise),
            window=window,
            window_agg=window_agg,
            diff_fn=diff_fn,
        )

    def __str__(self):
        return "PyODScorer (model {})".format(self.model.__str__().split("(")[0])

    def _model_score_method(self, model, data: np.ndarray) -> np.ndarray:
        """Wrapper around model inference method"""
        return model.decision_function(data)



from collections.abc import Mapping, Sequence
from typing import Any, Union

import numpy as np
import pytest

from darts import TimeSeries
from darts.dataprocessing import Pipeline
from darts.dataprocessing.transformers import (
    BaseDataTransformer,
    FittableDataTransformer,
    InvertibleDataTransformer,
    InvertibleMapper,
    Mapper,
)
from darts.utils.timeseries_generation import constant_timeseries


class TestPipeline:
    class DataTransformerMock1(BaseDataTransformer):
        def __init__(self):
            super().__init__()
            self.transform_called = False
            self.inverse_transform_called = False
            self.fit_called = False

        @staticmethod
        def ts_transform(data: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
            return data.append_values(constant_timeseries(value=1, length=3).values())

        def transform(self, data, *args, **kwargs) -> TimeSeries:
            self.transform_called = True
            return super().transform(data, *args, **kwargs)

    class DataTransformerMock2(FittableDataTransformer, InvertibleDataTransformer):
        def __init__(self):
            super().__init__()

            self.transform_called = False
            self.inverse_transform_called = False
            self.fit_called = False

        @staticmethod
        def ts_fit(series: TimeSeries, params: Mapping[str, Any]):
            pass

        @staticmethod
        def ts_transform(series: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
            return series.append_values(constant_timeseries(value=2, length=3).values())

        @staticmethod
        def ts_inverse_transform(
            series: TimeSeries, params: Mapping[str, Any]
        ) -> TimeSeries:
            return series

        def fit(self, data):
            super().fit(data)
            self.fit_called = True
            return self

        def transform(self, data, *args, **kwargs):
            self.transform_called = True
            self.args = args
            self.kwargs = kwargs
            return super().transform(data, *args, **kwargs)

        def inverse_transform(self, data, *args, **kwargs) -> TimeSeries:
            self.inverse_transform_called = True
            self.args = args
            self.kwargs = kwargs
            return super().inverse_transform(data, *args, **kwargs)

    class PlusTenTransformer(InvertibleDataTransformer):
        def __init__(self, name="+10 transformer"):
            super().__init__(name=name)

        @staticmethod
        def ts_transform(series: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
            return series.map(lambda x: x + 10)

        @staticmethod
        def ts_inverse_transform(
            series: TimeSeries, params: Mapping[str, Any]
        ) -> TimeSeries:
            return series.map(lambda x: x - 10)

    class TimesTwoTransformer(InvertibleDataTransformer):
        def __init__(self):
            super().__init__(name="*2 transformer")

        @staticmethod
        def ts_transform(data: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
            return data.map(lambda x: x * 2)

        @staticmethod
        def ts_inverse_transform(
            data: TimeSeries, params: Mapping[str, Any]
        ) -> TimeSeries:
            return data.map(lambda x: x / 2)

    class ExtendTransformer(FittableDataTransformer, InvertibleDataTransformer):
        def __init__(self, global_fit: bool, coef: int):
            self.coef = coef
            super().__init__(
                name="fittable extending transformer", global_fit=global_fit
            )

        @staticmethod
        def ts_fit(
            series: Union[TimeSeries, Sequence[TimeSeries]],
            params: Mapping[str, Any],
            *args,
            **kwargs,
        ):
            coef = params["fixed"]["coef"]
            if isinstance(series, Sequence):
                return sum(ts.values()[0] for ts in series) + coef
            else:
                return series.values()[0] + coef

        @staticmethod
        def ts_transform(data: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
            return data + params["fitted"]

        @staticmethod
        def ts_inverse_transform(
            data: TimeSeries, params: Mapping[str, Any]
        ) -> TimeSeries:
            return data - params["fitted"]

    def test_transform(self):
        # given
        mock1 = self.DataTransformerMock1()
        mock2 = self.DataTransformerMock2()
        data = constant_timeseries(value=0, length=3)
        transformers = [mock1] * 10 + [mock2] * 10
        p = Pipeline(transformers)
        # when
        p.fit(data)
        transformed = p.transform(data)

        # then
        assert 63 == len(transformed)
        assert [0] * 3 + [1] * 30 + [2] * 30 == list(transformed.values())
        for t in transformers:
            assert t.transform_called
            assert not t.inverse_transform_called

    def test_transform_prefitted(self):
        """Check that when multiple series are passed to fit transformers with global_fit=False,
        transform behave as expected when series_idx is specified.

        Note: the transformers are fitted independently to make the expected results more intuitive
        """
        data = [
            constant_timeseries(value=0, length=2),
            constant_timeseries(value=10, length=2),
        ]

        def get_transf(global_fit: bool, fit: bool, coef: int):
            transf = self.ExtendTransformer(global_fit=global_fit, coef=coef)
            if fit:
                transf.fit(data)
            return transf

        # multiple series, global_fit=False
        p = Pipeline([
            get_transf(global_fit=False, fit=True, coef=1),
            get_transf(global_fit=False, fit=True, coef=5),
        ])
        transformed = p.transform(data)

        # ts + (data[0][0] + 1) + (data[0][0] + 5) = 6
        np.testing.assert_array_almost_equal(
            transformed[0].values(), np.array([[6, 6]]).T
        )
        # ts + (data[1][0] + 1) + (data[1][0] + 5) = 10 + 11 + 15 = 36
        np.testing.assert_array_almost_equal(
            transformed[1].values(), np.array([[36, 36]]).T
        )
        # implicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            transformed[0].values(), p.transform(data[0]).values()
        )
        # explicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            transformed[0].values(), p.transform(data[0], series_idx=[0]).values()
        )
        # implicitly use the first params of each transformer
        # ts + (data[0][0] + 1) + (data[0][0] + 5) = 10 + 1 + 5 = 16
        np.testing.assert_array_almost_equal(
            np.array([[16, 16]]).T, p.transform(data[1]).values()
        )
        # explicitly use the second params of each transformer
        np.testing.assert_array_almost_equal(
            transformed[1].values(), p.transform(data[1], series_idx=[1]).values()
        )

        # multiple series, mixture of local and global transformers
        p = Pipeline([
            get_transf(global_fit=False, fit=True, coef=1),
            get_transf(global_fit=True, fit=True, coef=90),
        ])
        transformed = p.transform(data)
        # ts + (data[0][0] + 1) + (sum(data[;, 0]) + 90) = 0 + 1 + 100
        np.testing.assert_array_almost_equal(
            transformed[0].values(), np.array([[101, 101]]).T
        )
        # ts + (data[1][0] + 1) + (sum(data[;, 0]) + 90) = 10 + 11 + 100
        np.testing.assert_array_almost_equal(
            transformed[1].values(), np.array([[121, 121]]).T
        )
        # implicitly use the first params of first transformer, the second is global
        np.testing.assert_array_almost_equal(
            transformed[0].values(), p.transform(data[0]).values()
        )
        # explicitly use the first params of first transformer, the second is global
        np.testing.assert_array_almost_equal(
            transformed[0].values(), p.transform(data[0], series_idx=[0]).values()
        )
        # implicitly use the first params of first transformer, the second is global
        # ts + (data[0][0] + 1) + (sum(data[;, 0]) + 90) = 10 + 1 + 100
        np.testing.assert_array_almost_equal(
            np.array([[111, 111]]).T, p.transform(data[1]).values()
        )
        # explicitly use the second params of first transformer, the second is global
        np.testing.assert_array_almost_equal(
            transformed[1].values(), p.transform(data[1], series_idx=[1]).values()
        )

        # reversing input, and explicitly selecting reversed indexes
        assert transformed[::-1] == p.transform(data[::-1], series_idx=[1, 0])

    def test_inverse_raise_exception(self):
        # given
        mock = self.DataTransformerMock1()
        p = Pipeline([mock])

        # when & then
        with pytest.raises(ValueError):
            p.inverse_transform(None)

    def test_transformers_not_modified(self):
        # given
        mock = self.DataTransformerMock1()
        p = Pipeline([mock], copy=True)

        # when
        p.transform(constant_timeseries(value=1, length=10))

        # then
        assert not mock.transform_called

    def test_fit(self):
        # given
        data = constant_timeseries(value=0, length=3)
        transformers = [self.DataTransformerMock1() for _ in range(10)] + [
            self.DataTransformerMock2() for _ in range(10)
        ]
        p = Pipeline(transformers)

        # when
        p.fit(data)

        # then
        for i in range(10):
            assert not transformers[i].fit_called
        for i in range(10, 20):
            assert transformers[i].fit_called

    def test_fit_skips_superfluous_transforms(self):
        # given
        data = constant_timeseries(value=0, length=100)
        transformers = (
            [self.DataTransformerMock1() for _ in range(10)]
            + [self.DataTransformerMock2()]
            + [self.DataTransformerMock1() for _ in range(10)]
        )
        p = Pipeline(transformers)

        # when
        p.fit(data)

        # then
        for i in range(10):
            assert transformers[i].transform_called
        assert transformers[10].fit_called
        assert not transformers[10].transform_called
        for i in range(11, 21):
            assert not transformers[i].transform_called

    def test_transform_fit(self):
        # given
        data = constant_timeseries(value=0, length=3)
        transformers = [self.DataTransformerMock1() for _ in range(10)] + [
            self.DataTransformerMock2() for _ in range(10)
        ]
        p = Pipeline(transformers)

        # when
        _ = p.fit_transform(data)

        # then
        for t in transformers:
            assert t.transform_called
        for i in range(10):
            assert not transformers[i].fit_called
        for i in range(10, 20):
            assert transformers[i].fit_called

    def test_inverse_transform(self):
        # given
        data = constant_timeseries(value=0.0, length=3)

        transformers = [self.PlusTenTransformer(), self.TimesTwoTransformer()]
        p = Pipeline(transformers)

        # when
        transformed = p.transform(data)
        back = p.inverse_transform(transformed)

        # then
        assert data == back

    def test_inverse_transform_prefitted(self):
        """Check that when multiple series are passed to fit transformers with global_fit=False,
        inverse_transform behave as expected when series_idx is specified.

        Note: the transformers are fitted independently to make the expected results more intuitive
        """
        data = [
            constant_timeseries(value=0, length=2),
            constant_timeseries(value=10, length=2),
        ]

        def get_transf(global_fit: bool, fit: bool, coef: int):
            transf = self.ExtendTransformer(global_fit=global_fit, coef=coef)
            if fit:
                transf.fit(data)
            return transf

        # multiple series, global_fit=False
        p = Pipeline([
            get_transf(global_fit=False, fit=True, coef=1),
            get_transf(global_fit=False, fit=True, coef=5),
        ])
        transformed = p.transform(data)

        # implicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            data[0].values(), p.inverse_transform(transformed[0]).values()
        )
        # explicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            data[0].values(),
            p.inverse_transform(transformed[0], series_idx=[0]).values(),
        )

        # 10 + 11 + 15
        np.testing.assert_array_almost_equal(
            np.array([[36, 36]]).T, transformed[1].values()
        )
        # implicitly use the first params of each transformer
        # inverse_transform[0][0] = lambda x: x - 1, inverse_transform[1][0] = lambda x: x - 5
        np.testing.assert_array_almost_equal(
            np.array([[30, 30]]).T, p.inverse_transform(transformed[1]).values()
        )
        np.testing.assert_array_almost_equal(
            np.array([[30, 30]]).T,
            p.inverse_transform(transformed[1], series_idx=0).values(),
        )
        # explicitly use the second params of each transformer
        # inverse_transform[0][0] = lambda x: x - 11, inverse_transform[1][0] = lambda x: x - 15
        np.testing.assert_array_almost_equal(
            data[1].values(), p.inverse_transform(transformed[1], series_idx=1).values()
        )

        # multiple series, mixture of local and global transformers
        p = Pipeline([
            get_transf(global_fit=False, fit=True, coef=1),
            get_transf(global_fit=True, fit=True, coef=90),
        ])
        transformed = p.transform(data)

        # implicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            data[0].values(), p.inverse_transform(transformed[0]).values()
        )
        # explicitly use the first params of each transformer
        np.testing.assert_array_almost_equal(
            data[0].values(),
            p.inverse_transform(transformed[0], series_idx=[0]).values(),
        )
        # 10 + 11 + 100
        np.testing.assert_array_almost_equal(
            np.array([[121, 121]]).T, transformed[1].values()
        )

        # implicitly use the first params of each transformer
        # inverse_transform[0][0] = lambda x: x - 1, inverse_transform = lambda x: x - 100
        np.testing.assert_array_almost_equal(
            np.array([[20, 20]]).T, p.inverse_transform(transformed[1]).values()
        )
        # explicitly use the second params of each transformer
        # inverse_transform[0][1] = lambda x: x - 11, inverse_transform = lambda x: x - 100
        np.testing.assert_array_almost_equal(
            data[1].values(),
            p.inverse_transform(transformed[1], series_idx=[1]).values(),
        )

        # reversing input, and explicitly selecting reversed indexes
        assert transformed[::-1] == p.transform(data[::-1], series_idx=[1, 0])

    def test_getitem(self):
        # given

        transformers = [
            self.PlusTenTransformer(name=f"+10 transformer{i}") for i in range(0, 10)
        ]
        p = Pipeline(transformers)

        # when & then
        # note : only compares string representations, since __getitem__() copies the transformers
        assert str(p[1]._transformers) == str([transformers[1]])
        assert str(p[4:8]._transformers) == str(transformers[4:8])

        with pytest.raises(ValueError):
            p["invalid attempt"]

    def test_raises_on_non_transformers(self):
        # given
        input_list = list(range(10))

        # when & then
        with pytest.raises(ValueError) as err:
            Pipeline(input_list)

        assert (
            str(err.value)
            == "transformers should be objects deriving from BaseDataTransformer"
        )

    def test_raises_on_bad_key(self):
        # given
        bad_key = 12.0
        p = Pipeline([])

        # when & then
        with pytest.raises(ValueError) as err:
            p[bad_key]
        assert str(err.value) == "key must be either an int or a slice"

    def test_multi_ts(self):
        series1 = constant_timeseries(value=0.0, length=3)
        series2 = constant_timeseries(value=1.0, length=3)

        data = [series1, series2]

        mapper1 = InvertibleMapper(fn=lambda x: x + 10, inverse_fn=lambda x: x - 10)
        mapper2 = InvertibleMapper(fn=lambda x: x * 10, inverse_fn=lambda x: x / 10)

        transformers = [mapper1, mapper2]
        p = Pipeline(transformers)

        # when
        transformed = p.transform(data)
        back = p.inverse_transform(transformed)

        # then
        assert data == back

    def test_pipeline_partial_inverse(self):
        series = constant_timeseries(value=0.0, length=3)

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten)
        mapper_inv = InvertibleMapper(fn=lambda x: x + 2, inverse_fn=lambda x: x - 2)

        series_plus_ten = mapper.transform(series)

        pipeline = Pipeline([mapper, mapper_inv])

        transformed = pipeline.transform(series)

        # should fail, since partial is False by default
        with pytest.raises(ValueError):
            pipeline.inverse_transform(transformed)

        back = pipeline.inverse_transform(transformed, partial=True)

        # while the +/- 2 is inverted, the +10 operation is not
        assert series_plus_ten == back

    def test_pipeline_verbose(self):
        """
        Checks if the verbose param applied to the pipeline is changing the verbosity level in the
        contained transformers.
        """

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten, verbose=True)
        mapper_inv = InvertibleMapper(
            fn=lambda x: x + 2, inverse_fn=lambda x: x - 2, verbose=True
        )

        verbose_value = False
        pipeline = Pipeline([mapper, mapper_inv], verbose=verbose_value)

        for transformer in pipeline:
            assert transformer._verbose == verbose_value

    def test_pipeline_n_jobs(self):
        """
        Checks if the n_jobs param applied to the pipeline is changing the verbosity level in the
        contained transformers.
        """

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten, n_jobs=1)
        mapper_inv = InvertibleMapper(
            fn=lambda x: x + 2, inverse_fn=lambda x: x - 2, n_jobs=2
        )

        n_jobs_value = -1
        pipeline = Pipeline([mapper, mapper_inv], n_jobs=n_jobs_value)

        for transformer in pipeline:
            assert transformer._n_jobs == n_jobs_value
